% Copyright (C) 2014  Daniel Filkovic

% This file is part of APAME-WT.

% APAME-WT is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.

% APAME-WT is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.

% You should have received a copy of the GNU General Public License
% along with APAME-WT.  If not, see <http://www.gnu.org/licenses/>.

% file grid_calculation.m

function [x, y, z, NBody, N1Body, N1, N, M, numPan, ...
          maxN, maxM, maxN1, maxM1, maxNBody, ...
          FF, S, ...
          u1, u2, u3, ...
          o1, o2, o3, ...
          n1, n2, n3, ...
          p1, p2, p3, ...
          cx, cy, cz, ...
          x1, x2, x3, x4, ...
          y1, y2, y3, y4, ...
          d1, d2, d3, d4] = grid_calculation(x, y, z, M1, N1, ...
                                             lift, omega, rotCenter, ...
                                             V, Vx, Vy, Vz, ...
                                             numWake, wakeLength, divWake, farfieldfaktor)

% get geometry matrix size
[maxN1, maxM1, numComp] = size(x);
maxM = maxM1-1;

% enlarge maximum number of panels in row direction by number of wake panels
maxN1 = 0;
maxNBody = 0;
for i=1:numComp
    if lift(numComp) == 1
        if N1(numComp)+numWake(numComp) > maxN1
            maxN1 = N1(numComp)+numWake(numComp);
        end
    else
        if N1(numComp) > maxN1
            maxN1 = N1(numComp);
        end
    end
    if N1(numComp) > maxNBody
        maxNBody = N1(numComp);
    end
end
maxN = maxN1-1;
maxNBody = maxNBody-1;

% extract some panel numbers
M = zeros(numComp, 1);
N = zeros(numComp, 1);
O = zeros(numComp, 1);
numPan = 0;
N1Body = zeros(numComp, 1);
NBody = zeros(numComp, 1);
for compIndex = 1:numComp
    % enlarge components number of rows by number of wake panels
    N1Body(compIndex) = N1(compIndex);
    if lift(compIndex) == 1
        N1(compIndex) = N1(compIndex)+numWake(compIndex);
    end
    M(compIndex) = M1(compIndex)-1;
    N(compIndex) = N1(compIndex)-1;
    NBody(compIndex) = N1Body(compIndex)-1;
    O(compIndex) = M(compIndex)*NBody(compIndex);
    numPan = numPan + O(compIndex);
end

% allocate grid information matrices
FF = zeros(maxN, maxM, numComp);
S  = zeros(maxN, maxM, numComp);
u1 = zeros(maxN, maxM, numComp);
u2 = zeros(maxN, maxM, numComp);
u3 = zeros(maxN, maxM, numComp);
o1 = zeros(maxN, maxM, numComp);
o2 = zeros(maxN, maxM, numComp);
o3 = zeros(maxN, maxM, numComp);
n1 = zeros(maxN, maxM, numComp);
n2 = zeros(maxN, maxM, numComp);
n3 = zeros(maxN, maxM, numComp);
p1 = zeros(maxN, maxM, numComp);
p2 = zeros(maxN, maxM, numComp);
p3 = zeros(maxN, maxM, numComp);
cx = zeros(maxN, maxM, numComp);
cy = zeros(maxN, maxM, numComp);
cz = zeros(maxN, maxM, numComp);
x1 = zeros(maxN, maxM, numComp);
y1 = zeros(maxN, maxM, numComp);
x2 = zeros(maxN, maxM, numComp);
y2 = zeros(maxN, maxM, numComp);
x3 = zeros(maxN, maxM, numComp);
y3 = zeros(maxN, maxM, numComp);
x4 = zeros(maxN, maxM, numComp);
y4 = zeros(maxN, maxM, numComp);
d1 = zeros(maxN, maxM, numComp);
d2 = zeros(maxN, maxM, numComp);
d3 = zeros(maxN, maxM, numComp);
d4 = zeros(maxN, maxM, numComp);

% go through each component at a time
for compIndex = 1:numComp
    % extend component geometry with wake only if lifting component
    if lift(compIndex) == 1
        % check if rotational or non-rotational body
        if omega(compIndex) == 0
            % calculate single panel extent
            extent = wakeLength(compIndex)/numWake(compIndex);
            
            % calculate extend direction based on velocity direction
            dispX = extent*Vx/V;
            dispY = extent*Vy/V;
            dispZ = extent*Vz/V;
            
            % calculate wake coordinates
            for colIndex = 1:M1(compIndex)
                wakeIndex = 1;
                for rowIndex = N1Body(compIndex)+1:N1(compIndex)
                    % extend coordinates linearly
                    x(rowIndex, colIndex, compIndex) = x(N1Body(compIndex), colIndex, compIndex)+dispX*wakeIndex;
                    y(rowIndex, colIndex, compIndex) = y(N1Body(compIndex), colIndex, compIndex)+dispY*wakeIndex;
                    z(rowIndex, colIndex, compIndex) = z(N1Body(compIndex), colIndex, compIndex)+dispZ*wakeIndex;
                    wakeIndex = wakeIndex+1;
                end
            end
        else
            % due to angular velocity of omega(compIndex) and angular
            % division of divWake, calculate time for singe wake step
            timeStep = 2*pi/omega(compIndex)/divWake(compIndex);
            
            % calculate single angle step
            angleStep = 2*pi/divWake(compIndex);
            
            % due to timeStep time and velocity components x and y
            % calculate their displacements
            dispX = timeStep*Vx;
            dispY = timeStep*Vy;
            dispZ = timeStep*Vz;

            % calculate wake coordinates
            for colIndex = 1:M1(compIndex)
                % find original polar coordinates
                yRel = y(N1Body(compIndex), colIndex, compIndex) - rotCenter(compIndex, 2);
                zRel = z(N1Body(compIndex), colIndex, compIndex) - rotCenter(compIndex, 3);
                radius = sqrt (yRel^2+zRel^2);
                phi    = atan2(zRel,  yRel  );
                wakeIndex = 1;
                for rowIndex = N1Body(compIndex)+1:N1(compIndex)
                    % extend x by x velocity component
                    x(rowIndex, colIndex, compIndex) = x(N1Body(compIndex), colIndex, compIndex)+dispX*wakeIndex;

                    % calculate new polar coordinates (new angle)
                    if omega(compIndex) == 0
                        phiMod = phi;
                    else
                        phiMod = phi+angleStep*wakeIndex;
                    end

                    % convert back to cartesian coordinate system (add y sidewind and z upwind displacement)
                    y(rowIndex, colIndex, compIndex) = radius*cos(phiMod)+dispY*wakeIndex + rotCenter(compIndex, 2);
                    z(rowIndex, colIndex, compIndex) = radius*sin(phiMod)+dispZ*wakeIndex + rotCenter(compIndex, 3);

                    wakeIndex = wakeIndex+1;
                end
            end
        end
    end
    
    % calculation of panel surfaces and surface normals
    for colIndex = 1:M
        for rowIndex = 1:N
            % calculate panel diagonals
            A = [(x(rowIndex+1, colIndex+1, compIndex) - x(rowIndex, colIndex  , compIndex)), ...
                 (y(rowIndex+1, colIndex+1, compIndex) - y(rowIndex, colIndex  , compIndex)), ...
                 (z(rowIndex+1, colIndex+1, compIndex) - z(rowIndex, colIndex  , compIndex))];
            B = [(x(rowIndex+1, colIndex  , compIndex) - x(rowIndex, colIndex+1, compIndex)), ...
                 (y(rowIndex+1, colIndex  , compIndex) - y(rowIndex, colIndex+1, compIndex)), ...
                 (z(rowIndex+1, colIndex  , compIndex) - z(rowIndex, colIndex+1, compIndex))];
             
            % and their magnitudes
            AA = realsqrt(A(1)^2 + A(2)^2 + A(3)^2);
            BB = realsqrt(B(1)^2 + B(2)^2 + B(3)^2);
            
            % calculate far field distance (far field factor multiplied by larger diagonal)
            FF(rowIndex, colIndex, compIndex) = farfieldfaktor*max(AA, BB);
            
            % calculate cross product of diagonals
            C = cross(A, B);
            
            % its modulus
            modc = realsqrt(C(1)^2 + C(2)^2 + C(3)^2);
            
            % surface area is half of cross product of diagonals
            S(rowIndex, colIndex, compIndex) = modc/2;
            
            % check if zero modulus (colinear diagonals)
            if modc == 0
                % nulify normal components
                n1(rowIndex, colIndex, compIndex) = 0;
                n2(rowIndex, colIndex, compIndex) = 0;
                n3(rowIndex, colIndex, compIndex) = 0;
            else
                % normalize cross product with modulus (get unit vector)
                norm = -C./modc;
                n1(rowIndex, colIndex, compIndex) = norm(1);
                n2(rowIndex, colIndex, compIndex) = norm(2);
                n3(rowIndex, colIndex, compIndex) = norm(3);
            end
        end
    end

    % calculation of colocation points
    for colIndex = 1:M
        for rowIndex = 1:N
            cx(rowIndex, colIndex, compIndex) = (x(rowIndex,   colIndex  , compIndex) + ...
                                                 x(rowIndex,   colIndex+1, compIndex) + ...
                                                 x(rowIndex+1, colIndex  , compIndex) + ...
                                                 x(rowIndex+1, colIndex+1, compIndex) )/4;
            cy(rowIndex, colIndex, compIndex) = (y(rowIndex,   colIndex  , compIndex) + ...
                                                 y(rowIndex,   colIndex+1, compIndex) + ...
                                                 y(rowIndex+1, colIndex  , compIndex) + ...
                                                 y(rowIndex+1, colIndex+1, compIndex) )/4;
            cz(rowIndex, colIndex, compIndex) = (z(rowIndex,   colIndex  , compIndex) + ...
                                                 z(rowIndex,   colIndex+1, compIndex) + ...
                                                 z(rowIndex+1, colIndex  , compIndex) + ...
                                                 z(rowIndex+1, colIndex+1, compIndex) )/4;
        end
    end

    % calculation of u(longitudinal), p(transversal) and o(perpendicular)
    % unit vectors
    for colIndex = 1:M
        for rowIndex = 1:N
            ux = ((x(rowIndex+1, colIndex, compIndex) + x(rowIndex+1, colIndex+1, compIndex)) - ...
                  (x(rowIndex  , colIndex, compIndex) + x(rowIndex,   colIndex+1, compIndex)) )/2;
            uy = ((y(rowIndex+1, colIndex, compIndex) + y(rowIndex+1, colIndex+1, compIndex)) - ...
                  (y(rowIndex  , colIndex, compIndex) + y(rowIndex,   colIndex+1, compIndex)) )/2;
            uz = ((z(rowIndex+1, colIndex, compIndex) + z(rowIndex+1, colIndex+1, compIndex)) - ...
                  (z(rowIndex,   colIndex, compIndex) + z(rowIndex,   colIndex+1, compIndex)) )/2;
            uu = realsqrt(ux^2 + uy^2 + uz^2);
            % avoid division by zero
            if uu == 0
                u1(rowIndex, colIndex, compIndex) = 0;
                u2(rowIndex, colIndex, compIndex) = 0;
                u3(rowIndex, colIndex, compIndex) = 0;
            else
                u1(rowIndex, colIndex, compIndex) = ux/uu;
                u2(rowIndex, colIndex, compIndex) = uy/uu;
                u3(rowIndex, colIndex, compIndex) = uz/uu;
            end
            px = ((x(rowIndex, colIndex+1, compIndex) + x(rowIndex+1, colIndex+1, compIndex)) - ...
                  (x(rowIndex, colIndex  , compIndex) + x(rowIndex+1, colIndex  , compIndex)) )/2;
            py = ((y(rowIndex, colIndex+1, compIndex) + y(rowIndex+1, colIndex+1, compIndex)) - ...
                  (y(rowIndex, colIndex  , compIndex) + y(rowIndex+1, colIndex  , compIndex)) )/2;
            pz = ((z(rowIndex, colIndex+1, compIndex) + z(rowIndex+1, colIndex+1, compIndex)) - ...
                  (z(rowIndex, colIndex  , compIndex) + z(rowIndex+1, colIndex  , compIndex)) )/2;
            pp = sqrt(px^2 + py^2 + pz^2);
            % avoid division by zero
            if pp == 0
                p1(rowIndex, colIndex, compIndex) = 0;
                p2(rowIndex, colIndex, compIndex) = 0;
                p3(rowIndex, colIndex, compIndex) = 0;
            else
                p1(rowIndex, colIndex, compIndex) = px/pp;
                p2(rowIndex, colIndex, compIndex) = py/pp;
                p3(rowIndex, colIndex, compIndex) = pz/pp;
            end
            o1(rowIndex, colIndex, compIndex) = ...
                n2(rowIndex, colIndex, compIndex)*u3(rowIndex, colIndex, compIndex) - ...
                n3(rowIndex, colIndex, compIndex)*u2(rowIndex, colIndex, compIndex) ;
            o2(rowIndex, colIndex, compIndex) = ...
                u1(rowIndex, colIndex, compIndex)*n3(rowIndex, colIndex, compIndex) - ...
                n1(rowIndex, colIndex, compIndex)*u3(rowIndex, colIndex, compIndex) ;
            o3(rowIndex, colIndex, compIndex) = ...
                n1(rowIndex, colIndex, compIndex)*u2(rowIndex, colIndex, compIndex) - ...
                n2(rowIndex, colIndex, compIndex)*u1(rowIndex, colIndex, compIndex) ;
        end
    end

    % calculation of panel coordinates (in panel CS relative to panel centers) and panel side lengths
    for colIndex = 1:M
        for rowIndex = 1:N
            x1(rowIndex, colIndex, compIndex) = ...
                (x(rowIndex,   colIndex,   compIndex)-cx(rowIndex, colIndex, compIndex))*u1(rowIndex, colIndex, compIndex) + ...
                (y(rowIndex,   colIndex,   compIndex)-cy(rowIndex, colIndex, compIndex))*u2(rowIndex, colIndex, compIndex) + ...
                (z(rowIndex,   colIndex,   compIndex)-cz(rowIndex, colIndex, compIndex))*u3(rowIndex, colIndex, compIndex) ;
            y1(rowIndex, colIndex, compIndex) = ...
                (x(rowIndex,   colIndex,   compIndex)-cx(rowIndex, colIndex, compIndex))*o1(rowIndex, colIndex, compIndex) + ...
                (y(rowIndex,   colIndex,   compIndex)-cy(rowIndex, colIndex, compIndex))*o2(rowIndex, colIndex, compIndex) + ...
                (z(rowIndex,   colIndex,   compIndex)-cz(rowIndex, colIndex, compIndex))*o3(rowIndex, colIndex, compIndex) ;
            x2(rowIndex, colIndex, compIndex) = ...
                (x(rowIndex+1, colIndex,   compIndex)-cx(rowIndex, colIndex, compIndex))*u1(rowIndex, colIndex, compIndex) + ...
                (y(rowIndex+1, colIndex,   compIndex)-cy(rowIndex, colIndex, compIndex))*u2(rowIndex, colIndex, compIndex) + ...
                (z(rowIndex+1, colIndex,   compIndex)-cz(rowIndex, colIndex, compIndex))*u3(rowIndex, colIndex, compIndex) ;
            y2(rowIndex, colIndex, compIndex) = ...
                (x(rowIndex+1, colIndex,   compIndex)-cx(rowIndex, colIndex, compIndex))*o1(rowIndex, colIndex, compIndex) + ...
                (y(rowIndex+1, colIndex,   compIndex)-cy(rowIndex, colIndex, compIndex))*o2(rowIndex, colIndex, compIndex) + ...
                (z(rowIndex+1, colIndex,   compIndex)-cz(rowIndex, colIndex, compIndex))*o3(rowIndex, colIndex, compIndex) ;
            x3(rowIndex, colIndex, compIndex) = ...
                (x(rowIndex+1, colIndex+1, compIndex)-cx(rowIndex, colIndex, compIndex))*u1(rowIndex, colIndex, compIndex) + ...
                (y(rowIndex+1, colIndex+1, compIndex)-cy(rowIndex, colIndex, compIndex))*u2(rowIndex, colIndex, compIndex) + ...
                (z(rowIndex+1, colIndex+1, compIndex)-cz(rowIndex, colIndex, compIndex))*u3(rowIndex, colIndex, compIndex) ;
            y3(rowIndex, colIndex, compIndex) = ...
                (x(rowIndex+1, colIndex+1, compIndex)-cx(rowIndex, colIndex, compIndex))*o1(rowIndex, colIndex, compIndex)+ ...
                (y(rowIndex+1, colIndex+1, compIndex)-cy(rowIndex, colIndex, compIndex))*o2(rowIndex, colIndex, compIndex)+ ...
                (z(rowIndex+1, colIndex+1, compIndex)-cz(rowIndex, colIndex, compIndex))*o3(rowIndex, colIndex, compIndex);
            x4(rowIndex, colIndex, compIndex) = ...
                (x(rowIndex,   colIndex+1, compIndex)-cx(rowIndex, colIndex, compIndex))*u1(rowIndex, colIndex, compIndex)+ ...
                (y(rowIndex,   colIndex+1, compIndex)-cy(rowIndex, colIndex, compIndex))*u2(rowIndex, colIndex, compIndex)+ ...
                (z(rowIndex,   colIndex+1, compIndex)-cz(rowIndex, colIndex, compIndex))*u3(rowIndex, colIndex, compIndex);
            y4(rowIndex, colIndex, compIndex) = ...
                (x(rowIndex,   colIndex+1, compIndex)-cx(rowIndex, colIndex, compIndex))*o1(rowIndex, colIndex, compIndex)+ ...
                (y(rowIndex,   colIndex+1, compIndex)-cy(rowIndex, colIndex, compIndex))*o2(rowIndex, colIndex, compIndex)+ ...
                (z(rowIndex,   colIndex+1, compIndex)-cz(rowIndex, colIndex, compIndex))*o3(rowIndex, colIndex, compIndex);
            d1(rowIndex, colIndex, compIndex) = ...
                realsqrt((x2(rowIndex, colIndex, compIndex)-x1(rowIndex, colIndex, compIndex))^2 + ...
                         (y2(rowIndex, colIndex, compIndex)-y1(rowIndex, colIndex, compIndex))^2 );
            d2(rowIndex, colIndex, compIndex) = ...
                realsqrt((x3(rowIndex, colIndex, compIndex)-x2(rowIndex, colIndex, compIndex))^2 + ...
                         (y3(rowIndex, colIndex, compIndex)-y2(rowIndex, colIndex, compIndex))^2);
            d3(rowIndex, colIndex, compIndex) = ...
                realsqrt((x4(rowIndex, colIndex, compIndex)-x3(rowIndex, colIndex, compIndex))^2 + ...
                         (y4(rowIndex, colIndex, compIndex)-y3(rowIndex, colIndex, compIndex))^2);
            d4(rowIndex, colIndex, compIndex) = ...
                realsqrt((x1(rowIndex, colIndex, compIndex)-x4(rowIndex, colIndex, compIndex))^2 + ...
                         (y1(rowIndex, colIndex, compIndex)-y4(rowIndex, colIndex, compIndex))^2);
        end
    end
end
