% Copyright (C) 2014  Daniel Filkovic

% This file is part of APAME-WT.

% APAME-WT is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.

% APAME-WT is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.

% You should have received a copy of the GNU General Public License
% along with APAME-WT.  If not, see <http://www.gnu.org/licenses/>.

% file post_calculation.m

function [v, cp, ...
          vx, vy, vz, ...
          CX, CY, CZ, ...
          CL, CM, CN, ...
          CX_comp, CY_comp, CZ_comp, ...
          CL_comp, CM_comp, CN_comp] = post_calculation(gama, ...
                                         c, span, Sw, V, Vy, Vz, VA, VB, VC, ro, ...
                                         omega, rotCenter, rotAxis, ...
                                         S, ...
                                         numComp, M, NBody, maxM, maxNBody, ...
                                         u1,u2,u3, ...
                                         n1,n2,n3, ...
                                         p1,p2,p3, ...
                                         o1,o2,o3, ...
                                         cx,cy,cz)

% allocate result matrices
FX = 0;
FY = 0;
FZ = 0;
FL = 0;
FM = 0;
FN = 0;
FX_comp = zeros(numComp, 1);
FY_comp = zeros(numComp, 1);
FZ_comp = zeros(numComp, 1);
FL_comp = zeros(numComp, 1);
FM_comp = zeros(numComp, 1);
FN_comp = zeros(numComp, 1);
vx = zeros(maxNBody, maxM, numComp);
vy = zeros(maxNBody, maxM, numComp);
vz = zeros(maxNBody, maxM, numComp);
v =  zeros(maxNBody, maxM, numComp);
cp = zeros(maxNBody, maxM, numComp);
q = 0.5*ro*V^2;
for compIndex = 1:numComp
    for colIndex = 1:M(compIndex)
        for rowIndex = 1:NBody(compIndex)
            % calculate local velocity component
            VxLocal = VA+VB*cz(rowIndex, colIndex, compIndex)+VC*cz(rowIndex, colIndex, compIndex)^2;
            
            % calculation of induced speeds in row direction
            if rowIndex == 1
                % boundary panels; first row - first order interpolation
                dx = cx(rowIndex+1, colIndex, compIndex) - cx(rowIndex, colIndex, compIndex);
                dy = cy(rowIndex+1, colIndex, compIndex) - cy(rowIndex, colIndex, compIndex);
                dz = cz(rowIndex+1, colIndex, compIndex) - cz(rowIndex, colIndex, compIndex);
                dr = sqrt(dx^2+dy^2+dz^2);
                qu = (gama(rowIndex+1, colIndex, compIndex) - gama(rowIndex, colIndex, compIndex))/dr;
            elseif rowIndex == NBody(compIndex)
                % boundary panels; last row - first order interpolation
                dx = cx(rowIndex, colIndex, compIndex) - cx(rowIndex-1, colIndex, compIndex);
                dy = cy(rowIndex, colIndex, compIndex) - cy(rowIndex-1, colIndex, compIndex);
                dz = cz(rowIndex, colIndex, compIndex) - cz(rowIndex-1, colIndex, compIndex);
                dr = sqrt(dx^2+dy^2+dz^2);
                qu = (gama(rowIndex, colIndex, compIndex) - gama(rowIndex-1, colIndex, compIndex))/dr;
            else
                % inner panel rows - second order interpolation
                dx = cx(rowIndex+1, colIndex, compIndex) - cx(rowIndex, colIndex, compIndex);
                dy = cy(rowIndex+1, colIndex, compIndex) - cy(rowIndex, colIndex, compIndex);
                dz = cz(rowIndex+1, colIndex, compIndex) - cz(rowIndex, colIndex, compIndex);
                drupwind = sqrt(dx^2+dy^2+dz^2);
                dx = cx(rowIndex, colIndex, compIndex) - cx(rowIndex-1, colIndex, compIndex);
                dy = cy(rowIndex, colIndex, compIndex) - cy(rowIndex-1, colIndex, compIndex);
                dz = cz(rowIndex, colIndex, compIndex) - cz(rowIndex-1, colIndex, compIndex);
                drdownwind = sqrt(dx^2+dy^2+dz^2);
                pu = polyfit([-drdownwind, 0, drupwind], ...
                             [gama(rowIndex-1, colIndex, compIndex), ...
                              gama(rowIndex,   colIndex, compIndex), ...
                              gama(rowIndex+1, colIndex, compIndex)], 2);
                qu = pu(2);
            end
            % calculation of induced speeds in column direction
            if colIndex == 1
                % boundary panels; first column - first order interpolation
                dx = cx(rowIndex, colIndex+1, compIndex) - cx(rowIndex, colIndex, compIndex);
                dy = cy(rowIndex, colIndex+1, compIndex) - cy(rowIndex, colIndex, compIndex);
                dz = cz(rowIndex, colIndex+1, compIndex) - cz(rowIndex, colIndex, compIndex);
                dr = sqrt(dx^2+dy^2+dz^2);
                qp = (gama(rowIndex, colIndex+1, compIndex) - gama(rowIndex, colIndex, compIndex))/dr;
            elseif colIndex == M(compIndex)
                % boundary panels; last column - first order interpolation
                dx = cx(rowIndex, colIndex, compIndex) - cx(rowIndex, colIndex-1, compIndex);
                dy = cy(rowIndex, colIndex, compIndex) - cy(rowIndex, colIndex-1, compIndex);
                dz = cz(rowIndex, colIndex, compIndex) - cz(rowIndex, colIndex-1, compIndex);
                dr = sqrt(dx^2+dy^2+dz^2);
                qp = (gama(rowIndex, colIndex, compIndex) - gama(rowIndex, colIndex-1, compIndex))/dr;
            else
                % inner panel columns - second order interpolation
                dx = cx(rowIndex, colIndex+1, compIndex) - cx(rowIndex, colIndex, compIndex);
                dy = cy(rowIndex, colIndex+1, compIndex) - cy(rowIndex, colIndex, compIndex);
                dz = cz(rowIndex, colIndex+1, compIndex) - cz(rowIndex, colIndex, compIndex);
                drupwind = sqrt(dx^2+dy^2+dz^2);
                dx = cx(rowIndex, colIndex, compIndex) - cx(rowIndex, colIndex-1, compIndex);
                dy = cy(rowIndex, colIndex, compIndex) - cy(rowIndex, colIndex-1, compIndex);
                dz = cz(rowIndex, colIndex, compIndex) - cz(rowIndex, colIndex-1, compIndex);
                drdownwind = sqrt(dx^2+dy^2+dz^2);
                pp = polyfit([-drdownwind, 0, drupwind], ...
                             [gama(rowIndex, colIndex-1, compIndex), ...
                              gama(rowIndex, colIndex,   compIndex), ...
                              gama(rowIndex, colIndex+1, compIndex)], 2);
                qp = pp(2);
            end
            
            qo = (p1(rowIndex, colIndex, compIndex)*o1(rowIndex, colIndex, compIndex) + ...
                  p2(rowIndex, colIndex, compIndex)*o2(rowIndex, colIndex, compIndex) + ...
                  p3(rowIndex, colIndex, compIndex)*o3(rowIndex, colIndex, compIndex) )*qp;

            if (omega(compIndex) ~= 0)
                % subtract rotation influence on free stream velocity
                % velocity vector is cross-product of rotational axis vector (of
                % rotational velocity magnitude) and radius vector (vector from
                % rotational axis origin to panel center)
            
                % make sure that rotational axis unit vector is of unit magnitude
                rotAxisMagn = sqrt(rotAxis(compIndex, 1)^2 + ...
                                   rotAxis(compIndex, 2)^2 + ...
                                   rotAxis(compIndex, 3)^2 );
                if (rotAxisMagn ~= 1)
                    rotAxis(compIndex) = rotAxis(compIndex)/rotAxisMagn;
                end

                % multiply it by rotational velocity
                vecOmega = rotAxis(compIndex, :)*omega(compIndex);

                % get radius vector
                radVec(1) = cx(rowIndex, colIndex, compIndex) - rotCenter(compIndex, 1);
                radVec(2) = cy(rowIndex, colIndex, compIndex) - rotCenter(compIndex, 2);
                radVec(3) = cz(rowIndex, colIndex, compIndex) - rotCenter(compIndex, 3);

                % get cross product
                tangVel = cross(vecOmega, radVec);
                
                gu = u1(rowIndex, colIndex, compIndex)*(VxLocal-tangVel(1)) + ...
                     u2(rowIndex, colIndex, compIndex)*(Vy     -tangVel(2)) + ...
                     u3(rowIndex, colIndex, compIndex)*(Vz     -tangVel(3)) ;
                go = o1(rowIndex, colIndex, compIndex)*(VxLocal-tangVel(1)) + ...
                     o2(rowIndex, colIndex, compIndex)*(Vy     -tangVel(2)) + ...
                     o3(rowIndex, colIndex, compIndex)*(Vz     -tangVel(3)) ;
            else
                gu = u1(rowIndex, colIndex, compIndex)*VxLocal + ...
                     u2(rowIndex, colIndex, compIndex)*Vy      + ...
                     u3(rowIndex, colIndex, compIndex)*Vz      ;
                go = o1(rowIndex, colIndex, compIndex)*VxLocal + ...
                     o2(rowIndex, colIndex, compIndex)*Vy      + ...
                     o3(rowIndex, colIndex, compIndex)*Vz      ;
            end

            vx(rowIndex, colIndex, compIndex) = (-qu + gu)*u1(rowIndex, colIndex, compIndex) + ...
                                                (-qo + go)*o1(rowIndex, colIndex, compIndex) ;
            vy(rowIndex, colIndex, compIndex) = (-qu + gu)*u2(rowIndex, colIndex, compIndex) + ...
                                                (-qo + go)*o2(rowIndex, colIndex, compIndex) ;
            vz(rowIndex, colIndex, compIndex) = (-qu + gu)*u3(rowIndex, colIndex, compIndex) + ...
                                                (-qo + go)*o3(rowIndex, colIndex, compIndex) ;
            v (rowIndex, colIndex, compIndex) = realsqrt(vx(rowIndex, colIndex, compIndex)^2 + ...
                                                         vy(rowIndex, colIndex, compIndex)^2 + ...
                                                         vz(rowIndex, colIndex, compIndex)^2 );

            cp(rowIndex, colIndex, compIndex) = 1-v(rowIndex, colIndex, compIndex)^2/V^2;
            dX = -cp(rowIndex, colIndex, compIndex) * ...
                   S(rowIndex, colIndex, compIndex) * ...
                  n1(rowIndex, colIndex, compIndex) *q;
            dY = -cp(rowIndex, colIndex, compIndex) * ...
                   S(rowIndex, colIndex, compIndex) * ...
                  n2(rowIndex, colIndex, compIndex) *q;
            dZ = -cp(rowIndex, colIndex, compIndex) * ...
                   S(rowIndex, colIndex, compIndex) * ...
                  n3(rowIndex, colIndex, compIndex) *q;
            FX_comp(compIndex) = FX_comp(compIndex) + dX;
            FY_comp(compIndex) = FY_comp(compIndex) + dY;
            FZ_comp(compIndex) = FZ_comp(compIndex) + dZ;
            FL_comp(compIndex) = FL_comp(compIndex)                                        - dY*cz(rowIndex, colIndex, compIndex) + dZ*cy(rowIndex, colIndex, compIndex);
            FM_comp(compIndex) = FM_comp(compIndex) + dX*cz(rowIndex, colIndex, compIndex)                                        - dZ*cx(rowIndex, colIndex, compIndex);
            FN_comp(compIndex) = FN_comp(compIndex) - dX*cy(rowIndex, colIndex, compIndex) + dY*cx(rowIndex, colIndex, compIndex) ;             
        end
    end
    FX = FX + FX_comp(compIndex);
    FY = FY + FY_comp(compIndex);
    FZ = FZ + FZ_comp(compIndex);
    FL = FL + FL_comp(compIndex);
    FM = FM + FM_comp(compIndex);
    FN = FN + FN_comp(compIndex);
end

CX = FX/q/Sw;
CY = FY/q/Sw;
CZ = FZ/q/Sw;
CL = FL/q/Sw/span;
CM = FM/q/Sw/c;
CN = FN/q/Sw/span;

CX_comp = FX_comp/q/Sw;
CY_comp = FY_comp/q/Sw;
CZ_comp = FZ_comp/q/Sw;
CL_comp = FL_comp/q/Sw/span;
CM_comp = FM_comp/q/Sw/c;
CN_comp = FN_comp/q/Sw/span;
